/*  callbacks.c

    ImSafe - Host Based Intrusion Detection
    Version 0.2.2 - Alpha

    Copyright (C) 2001  Laurent Eschenauer 
    <laurent_e@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include <sys/ioctl.h>
#include <unistd.h>

#include "common.h"
#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "learning.h"

/* Global */
int gImmDevice;
gboolean gDeviceOpened=FALSE;
gboolean gTraceStarted=FALSE;
gboolean gSensorOpened=FALSE;
gboolean gMonitorConfigured=FALSE;

/* Complicated dialog boxes can be slow to create, so we create them once and
   keep pointers to them in static variables so we can reuse them. */

static GtkWidget *open_device_w = NULL;
static GtkWidget *connect_to_sensor_w = NULL;
static GtkWidget *open_filesel_w = NULL;
static GtkWidget *tree_w=NULL;
static GtkWidget *about_w=NULL;
static GtkWidget *learn_w=NULL;
static GtkWidget *configure_monitor_w=NULL;
static GtkWidget *configure_trace_w=NULL;
static GtkWidget *trace_recorder_w=NULL;

static gint sCallbackId;

/* A key used to store pointers to the main window. */
const gchar *MainWindowKey = "MainWindowKey";
static const gchar *TargetTextKey = "TargetTextKey";
static const gchar *TreeItemKey   = "TreeItemKey";

/* Prototypes */
gint Stop_trace(void);
gint CloseDevice (void);
void on_subitem_select(GtkWidget *,gpointer);


/**************************************************/
/* Graphic functions                              */
/* Later on, we should put those out of here.. !  */
/**************************************************/


void UpdateSegs (struct Segments *segs,float value)
{
  int i;
  int y2;
  GdkSegment *tmp1,*tmp2;

  y2=200.0*value;
  

  for (i=0;i<segs->nsegs;i++)
    {
      tmp1=segs->seg+i;
      tmp2=segs->seg+i+1;
      tmp1->y1=tmp1->y2;
      tmp1->y2=tmp2->y1;
    }
  
  (segs->seg+segs->nsegs)->y1=(segs->seg+segs->nsegs)->y2;
  (segs->seg+segs->nsegs)->y2=y2;

}

struct Segments *NewSegs (int nsegs)
{
  int i;
  struct Segments *tmp;
  GdkSegment *seg;

  tmp=(struct Segments *) malloc (sizeof(struct Segments));
  tmp->nsegs=nsegs;
  tmp->seg=(GdkSegment *) malloc ((nsegs+1)*sizeof(GdkSegment));

  for (i=0;i<nsegs;i++)
    {
      seg=tmp->seg+i;
      seg->x1=i;
      seg->x2=(i+1);
      seg->y1=200;
      seg->y2=200;
    }
  return tmp;
}





/******************************************************/
/* Our GTK Functions : Those are not created by GLADE */
/*                                                    */
/******************************************************/


void
Our_exit (gint error)
{
 
  if (gTraceStarted)
    Stop_trace();
  if (gDeviceOpened)
    CloseDevice();
  
  gtk_exit (0);
}


void HandleAlert(int syscall_count,int pid,GtkWidget *event_log) 
{
  GString *message;

  message = g_string_new(NULL);
  g_string_sprintf(message,"ALERT: High risk of intrusion in process %d\n",pid);

  gtk_text_insert(GTK_TEXT(event_log),
		  NULL, 
		  &gColorRed, 
		  NULL,
		  message->str,message->len);
  
  gLastAlert=syscall_count;
  
  if(gOptions->exec_script)
    {
      gtk_text_insert(GTK_TEXT(event_log),
		  NULL, 
		  &gColorRed, 
		  NULL,
		  "Executing alert script now...\n",-1);
      system(gOptions->script_path);
    }
  return;
}

void FBODAlert(int syscall,int syscall_count,int pid,GtkWidget *event_log) 
{
  GString *message;

  message = g_string_new(NULL);
  g_string_sprintf(message,"ALERT:  Buffer Overflow in process %d\n Illegal execve\n",pid);

  gtk_text_insert(GTK_TEXT(event_log),
		  NULL, 
		  &gColorRed, 
		  NULL,
		  message->str,message->len);
  
  if(gOptions->exec_script)
    {
      gtk_text_insert(GTK_TEXT(event_log),
		  NULL, 
		  &gColorRed, 
		  NULL,
		  "Executing alert script now...\n",-1);
      system(gOptions->script_path);
    }
  return;
  
 
  return;
}

gint
AnalyseTrace(gpointer data)
{

  GtkWidget *event_log,*main_window,*mean_bar,*drawingarea,*label;
  int SizeRead,i,nd;
  int syscall_count,pid,syscall;
  unsigned long eip;
  gfloat proba=1.0;
  int *current_seq;
  gboolean UpdateBar=FALSE;
  GString *message;

  event_log=(GtkWidget *)data;
  main_window= gtk_widget_get_toplevel (GTK_WIDGET(event_log));
  
  drawingarea= lookup_widget (main_window, "drawingarea");
  syscall_count=0;

  SizeRead=0;
  if (gDeviceOpened)
    {
      SizeRead = read(gImmDevice, gBuffer, gOptions->buffer_size*sizeof(struct event));
      ioctl(gImmDevice,IMM_BUFFER_RESET,0);
      SizeRead = SizeRead/sizeof(struct event);
    }
  else if (gSensorOpened)
    {
      for(i=0;i<gOptions->buffer_size;i++)
	{ 
	  nd=fscanf(gSensor->in,"%d %d %d\n",&gBuffer[i].count,&gBuffer[i].pid,&gBuffer[i].call);

	  if (nd !=3)
	    break;
	  SizeRead++;
	}
    }
  
  if (SizeRead >1)
    {
      i=0;
      while (i<SizeRead-1)
	{
	  syscall_count=gBuffer[i].count;
	  pid=gBuffer[i].pid;
	  syscall=gBuffer[i].call;
	  
	  if(gOptions->fbod && syscall==11)
	    {
	      eip=gBuffer[i].eip;
	      if(eip>gOptions->fbod_threshold_up || eip<gOptions->fbod_threshold_down )
		{
		  printf("EIP: 0x%x \n",(unsigned int) eip);
		  FBODAlert(syscall,syscall_count,pid,event_log);
		}
	    }
	  current_seq=GetActiveSeq(pid,syscall,&gActive_stream,gOptions->seq_len);
	  if (current_seq!=NULL)
	    {
	      ProcessProba(gTree,current_seq,&proba,gOptions->seq_len);
	      UpdateMean(gMean,proba);
	      UpdateSegs(gSegs,gMean->mean);
	      if (gOptions->log_to_file)
		fprintf(gLogFile,"%d \t%f\n",syscall_count,gMean->mean);
	      if (gMean->mean<gOptions->imm_threshold && (syscall_count-gLastAlert)>100)
		HandleAlert(syscall_count,pid,event_log);
	      UpdateBar=TRUE;
	    }
	  i+=1;
	};
    }
  
  
  if (UpdateBar)
    {
      mean_bar = lookup_widget (main_window, "mean_progressbar");
      gtk_progress_set_percentage (GTK_PROGRESS(mean_bar),gMean->mean);

      if (GTK_WIDGET_VISIBLE(drawingarea))
	gtk_signal_emit_by_name(GTK_OBJECT(drawingarea),"expose_event");

      label=lookup_widget (main_window, "total_call_label");
      message = g_string_new(NULL);
      g_string_sprintf(message,"%d",
		       syscall_count);
      gtk_label_set_text(GTK_LABEL(label),message->str);
    }
  return TRUE;
}
   
gint
RecordTrace(gpointer data)
{

  GtkWidget *event_log,*main_window,*label,*tracer_window;
  int SizeRead,i,nd;
  int syscall_count,pid,syscall;
  GString *message;
  label=(GtkWidget *)data;
  
  /* raff, esthetic mods only */
  syscall_count=0;

  tracer_window= gtk_widget_get_toplevel (GTK_WIDGET(label));
  main_window = gtk_object_get_data (GTK_OBJECT (tracer_window), MainWindowKey);
  event_log = lookup_widget (main_window, "text_event");
  
    SizeRead=0;
  if (gDeviceOpened)
    {
      SizeRead = read(gImmDevice, gBuffer, gOptions->buffer_size*sizeof(struct event));
      ioctl(gImmDevice,IMM_BUFFER_RESET,0);
      SizeRead = SizeRead/sizeof(struct event);
    }
  else if (gSensorOpened)
    {
      printf("Reading from sensor...\n");
      for(i=0;i<gOptions->buffer_size;i++)
	{
	  nd=fscanf(gSensor->in,"%d %d %d\n",&gBuffer[i].count,&gBuffer[i].pid,&gBuffer[i].call);
	  printf("Nd: %d --> <%d, %d, %d>\n",nd,gBuffer[i].count,gBuffer[i].pid,gBuffer[i].call);
	      
	  if (nd !=3)
	    break;
     
	  SizeRead++;
	}
    }

  if (SizeRead >1)
    {
      i=0;
      while (i<SizeRead-1)
	{
	  syscall_count=gBuffer[i].count;
	  pid=gBuffer[i].pid;
	  syscall=gBuffer[i].call;

	  fprintf(gLogFile,"%d %d %d\n",syscall_count,pid,syscall);
	    
	  i+=1;
	};

      label=lookup_widget (tracer_window, "trace_calls_label");
      message = g_string_new(NULL);
      g_string_sprintf(message,"%d",
		       syscall_count);
      gtk_label_set_text(GTK_LABEL(label),message->str);
    }
  return TRUE;
}
   

gint CloseDevice (void)
{
  if (gDeviceOpened){
    close(gImmDevice);
    gDeviceOpened=FALSE;
  }
  else if (gSensorOpened)
    {
      if (gSensor->in)
	fclose (gSensor->in); 
    }
  return(0);
}


void
Open_device(GtkWidget *open_device)
{
  GtkWidget *main_window,*event_log;
  GString *message;

  main_window = gtk_object_get_data (GTK_OBJECT (open_device), MainWindowKey);
  event_log = lookup_widget (main_window, "text_event");
  
  /* Tell the user we are trying to open the device */

  message = g_string_new(NULL);
  g_string_sprintf(message,"INFO: Opening device %s :",
		   gOptions->device_name);

  gtk_text_insert(GTK_TEXT(event_log),
		  NULL, 
		  &event_log->style->black, 
		  NULL,
		  message->str,message->len);

  /* We open the device /dev file */
  if((gImmDevice = open(gOptions->device_name, O_RDONLY, 0)) < 0)
    {
      /* Device ain't good, give an error and leave this.*/
      gtk_text_insert(GTK_TEXT(event_log),
		      NULL, 
		      &event_log->style->black, 
		      NULL,
		      "FAILED :-(\n",-1);
      return;
    }
  
  /* We opened the driver, tell the user it's a success */
  gtk_text_insert(GTK_TEXT(event_log),
		  NULL, 
		  &event_log->style->black, 
		  NULL,
		  "Success !\n",-1);
 
  /* Tell the user we are going to configure the device. */
  /* We show the entire config so it's easy to debug. */

  g_string_sprintf(message,"INFO: Configuring device.\n");
  
  gtk_text_insert(GTK_TEXT(event_log),
		  NULL, 
		  &event_log->style->black, 
		  NULL,
		  message->str,message->len);
  
  /* Set the process name to the device by ioctl */
  if(ioctl(gImmDevice, IMM_SET_NAME, gOptions->imm_process_name))
       {
	 gtk_text_insert(GTK_TEXT(event_log),
			 NULL,
			 &event_log->style->black,
			 NULL,
			 "ERROR: Unable to set process name\n",-1);
	 return;
       }

  /* Set if execve calls must be traced and followed */
   if (gOptions->imm_execve_trace)
     {
       if(ioctl(gImmDevice, IMM_EXECVE_TRACE_ON, 0))
	 {
	   gtk_text_insert(GTK_TEXT(event_log),
			 NULL,
			 &event_log->style->black,
			 NULL,
			 "ERROR: Unable to set execve_trace\n",-1);
	   return;
	 }
     }
   
   else
     if(ioctl(gImmDevice, IMM_EXECVE_TRACE_OFF, 0))
       {
	 gtk_text_insert(GTK_TEXT(event_log),
			 NULL,
			 &event_log->style->black,
			 NULL,
			 "ERROR: Unable to set execve_trace\n",-1);
	   return;
       }

   /* If we are here, it's a success ! */
   /* We have an opened device */
   gtk_text_insert(GTK_TEXT(event_log),
		   NULL, 
		   &event_log->style->black, 
		   NULL,
		   "INFO: Configure success !\n",-1);
   gDeviceOpened=TRUE;
}


void
Open_sensor(GtkWidget *open_device)
{
  GtkWidget *main_window,*event_log;
  GString *message;
  int error=0,slen;
 

  main_window = gtk_object_get_data (GTK_OBJECT (connect_to_sensor_w), MainWindowKey);
  event_log = lookup_widget (main_window, "text_event");
  
  /* Tell the user we are trying to open the sensor */

  message = g_string_new(NULL);
  g_string_sprintf(message,"INFO: Creating sensor socket %s \n",
		   gOptions->sensor_name);

  gtk_text_insert(GTK_TEXT(event_log),
		  NULL, 
		  &event_log->style->black, 
		  NULL,
		  message->str,message->len);

  if (gSensor==NULL)
    gSensor=(sensor*) malloc (sizeof(sensor));

  gSensor->saddr.sa_family = AF_UNIX;
  strcpy( gSensor->saddr.sa_data, gOptions->sensor_name);
  slen=sizeof(gSensor->saddr);
  unlink (gSensor->saddr.sa_data);

  gSensor->s = socket(AF_UNIX, SOCK_STREAM, 0);
  if (gSensor->s == -1) LOG_EVENT("Error creating socket\n");
  if (bind(gSensor->s, &(gSensor->saddr), slen) != 0) LOG_EVENT("Error binding socket");


   gtk_text_insert(GTK_TEXT(event_log),
		   NULL, 
		   &event_log->style->black, 
		   NULL,
		   "INFO: Configure of sensor is a success !\n",-1);

   gSensorOpened=TRUE;

}


void
Configure_monitor(GtkWidget *open_device)
{
  GtkWidget *main_window,*event_log,*label,*editable;
  GString *message;
  struct Tree_node *tmp;
  int total_seq;

  main_window = gtk_object_get_data (GTK_OBJECT (open_device), MainWindowKey);
  event_log = lookup_widget (main_window, "text_event");
  
  /* First we try to open the database and update our Tree. */
  if (LoadTree(gTree,gOptions->imm_db_path)<0)
    {
      gtk_text_insert(GTK_TEXT(event_log),
		      NULL, 
		      &event_log->style->black, 
		      NULL,
		      "ERROR: unable to open database.\n",-1);
      return;
    }

  UpdateProbas(gTree);
  gOptions->db_seq_len=gOptions->seq_len;
 
  editable = lookup_widget (main_window, "an_seq_len");
  message = g_string_new(NULL);
  g_string_sprintf(message,"%d",
		   gOptions->seq_len);
  
  gtk_entry_set_text(GTK_ENTRY(editable),message->str);
  
  /* Put the db_seq_len on screen */
  label=lookup_widget (main_window, "db_seq_len_label");
  message = g_string_new(NULL);
  g_string_sprintf(message,"%d",
		   gOptions->seq_len);
  gtk_label_set_text(GTK_LABEL(label),message->str);

  /* Put the total sequences on screen */
  tmp=gTree;
  total_seq=0;
  while (tmp->id!=-1)
    {
      total_seq+=tmp->count;
      tmp=tmp->brother;
    }

  /* Put the db_seq_len on screen */
  label=lookup_widget (main_window, "total_seq_label");
  message = g_string_new(NULL);
  g_string_sprintf(message,"%d",
		   total_seq);
  gtk_label_set_text(GTK_LABEL(label),message->str);
  
  /* We know the seq_len and prepare space for a stream */
  
  gActive_stream.pid=-1;
  gActive_stream.next=NULL;
  gActive_stream.in_seq=0;
  gActive_stream.seq=(int *) malloc(gOptions->seq_len*sizeof(int));

  /* Create place to dump the data */
  gBuffer=(struct event *)malloc(gOptions->buffer_size*sizeof(struct event));

  /* The device as been reset (syscall_count to 0), we reset the last alert.*/
  gLastAlert=0;

  /* Open and set up the log file if the user asked to log. */
  if (gOptions->log_to_file)
    if ((gLogFile= fopen(gOptions->log_path,"w"))==NULL)
      {
	gtk_text_insert(GTK_TEXT(event_log),
			NULL, 
			&event_log->style->black, 
			NULL,
			"ERROR: Cannot open log file -> log set to FALSE. !\n",-1);
	gOptions->log_to_file=0;
      }

 
   gMonitorConfigured=TRUE;
}


gint
Start_trace(GtkWidget *event_log)
{
  int rfd;

  if (gDeviceOpened)
    {
      if(ioctl(gImmDevice, IMM_START, 0))
	return(-1);
    }
  else
    {
      gtk_text_insert(GTK_TEXT(event_log),
		      NULL, 
		      &event_log->style->black, 
		      NULL,
		      "INFO: Waiting for connection from sensor. !\n",-1);

      gtk_widget_draw(GTK_WIDGET(event_log),NULL);
      
      if (listen(gSensor->s,1) !=0) 
	{
	  gtk_text_insert(GTK_TEXT(event_log),
			  NULL, 
			  &event_log->style->black, 
			  NULL,
			  "INFO: Error during connection from sensor. !\n",-1);
	  return(-1);
	}

      rfd = accept(gSensor->s, NULL, NULL);

      if (rfd == -1) 
	{
	  gtk_text_insert(GTK_TEXT(event_log),
			  NULL, 
			  &event_log->style->black, 
			  NULL,
			  "INFO: Error during connection from sensor. (accept) !\n",-1);
	  return(-1);
	}

      fcntl(rfd,F_SETFL,O_NONBLOCK);
      gSensor->in = fdopen (rfd, "r");
      
      if (gSensor->in == NULL) 
	{
	  gtk_text_insert(GTK_TEXT(event_log),
			  NULL, 
			  &event_log->style->black, 
			  NULL,
			  "INFO: Error during connection from sensor. (fopen) !\n",-1);
	  return(-1);
	}
    }

    
      gTraceStarted=TRUE;
      gStartTime=time(NULL);

      if (gMean!=NULL)
	free(gMean);

      gMean= NewMean (gOptions->mean_len);
  
  
  
  sCallbackId=gtk_timeout_add( gOptions->imm_callback,
			       AnalyseTrace,
			       event_log);

  printf("Starting tracing...\n");
  return(0);
}

gint
Stop_trace(void)
{
  if (gDeviceOpened)
    if(ioctl(gImmDevice, IMM_STOP, 0))
      return(-1);

  gTraceStarted=FALSE;
  gtk_timeout_remove(sCallbackId);

  if (gLogFile!=NULL)
    fclose(gLogFile);

  return(0);
}

gint
Start_trace_record(GtkWidget *label)
{
  GtkWidget *statusbar,*event_log,*main_window,*tracer_window;
  
  int rfd;

  tracer_window = gtk_widget_get_toplevel (GTK_WIDGET (label));
  main_window = gtk_object_get_data (GTK_OBJECT (tracer_window), MainWindowKey);
  statusbar = lookup_widget (GTK_WIDGET (main_window), "statusbar1");
  event_log = lookup_widget (GTK_WIDGET(statusbar), "text_event");
  
  if (gDeviceOpened)
    {
      if(ioctl(gImmDevice, IMM_START, 0))
	return(-1);
    }
  else
    {
      gtk_text_insert(GTK_TEXT(event_log),
		      NULL, 
		      &event_log->style->black, 
		      NULL,
		      "INFO: Waiting for connection from sensor. !\n",-1);
      
      if (listen(gSensor->s,1) !=0) 
	{
	  gtk_text_insert(GTK_TEXT(event_log),
			  NULL, 
			  &event_log->style->black, 
			  NULL,
			  "INFO: Error during connection from sensor. !\n",-1);
	  return(-1);
	}

      gtk_widget_draw(GTK_WIDGET(main_window),NULL);

      rfd = accept(gSensor->s, NULL, NULL);

      if (rfd == -1) 
	{
	  gtk_text_insert(GTK_TEXT(event_log),
			  NULL, 
			  &event_log->style->black, 
			  NULL,
			  "INFO: Error during connection from sensor. (accept) !\n",-1);
	  return(-1);
	}
      
      gtk_widget_draw(GTK_WIDGET(main_window),NULL);
	
      fcntl(rfd,F_SETFL,O_NONBLOCK);
      gSensor->in = fdopen (rfd, "r");
      
      if (gSensor->in == NULL) 
	{
	  gtk_text_insert(GTK_TEXT(event_log),
			  NULL, 
			  &event_log->style->black, 
			  NULL,
			  "INFO: Error during connection from sensor. (fopen) !\n",-1);
	  return(-1);
	}
    }


  gTraceStarted=TRUE;
  gStartTime=time(NULL);

  printf("Adding callback now..\n");
  sCallbackId=gtk_timeout_add( gOptions->imm_callback,
			       RecordTrace,
			       label);
  printf("Callback added now... \n");

  return(0);
}

gint
Stop_trace_record(void)
{

   if (gDeviceOpened)
     if(ioctl(gImmDevice, IMM_STOP, 0))
      return(-1);

  gTraceStarted=FALSE;
  gtk_timeout_remove(sCallbackId);

  if (gLogFile!=NULL)
    fclose(gLogFile);

  return(0);
}
/***********************************************************/
/* Raise window functions : Those are not created by GLDAE */
/*                                                         */
/* Maybe we could make one for all ? no                    */
/*                                                         */
/***********************************************************/



void
Raise_open_device_w (GtkWidget *main_window)
{
  /* We use the same file selection widget each time, so first
     of all we create it if it hasn't already been created. */
  if (open_device_w == NULL)
    open_device_w = create_open_device_w ();

  /* We save a pointer to the main window inside the file selection's
     data list, so we can get it easily in the callbacks. */
  gtk_object_set_data (GTK_OBJECT (open_device_w), MainWindowKey, main_window);

  gtk_widget_show (open_device_w);
  gdk_window_raise (open_device_w->window);
}

void
Raise_connect_to_sensor_w (GtkWidget *main_window)
{
  /* We use the same file selection widget each time, so first
     of all we create it if it hasn't already been created. */

  if (connect_to_sensor_w == NULL)
    connect_to_sensor_w = create_open_sensor ();

  /* We save a pointer to the main window inside the file selection's
     data list, so we can get it easily in the callbacks. */
  gtk_object_set_data (GTK_OBJECT (connect_to_sensor_w), MainWindowKey, main_window);

  gtk_widget_show (connect_to_sensor_w);
  gdk_window_raise (connect_to_sensor_w->window);
}


void
Raise_learn_w (GtkWidget *main_window)
{
  /* We use the learn window widget each time, so first
     of all we create it if it hasn't already been created. */

  if (learn_w == NULL)
    learn_w = create_learn_w ();

  /* We save a pointer to the main window inside the file selection's
     data list, so we can get it easily in the callbacks. */
  gtk_object_set_data (GTK_OBJECT (learn_w), MainWindowKey, main_window);

  gtk_widget_show (learn_w);
  gdk_window_raise (learn_w->window);
}


void
Raise_browse_file (GtkWidget *main_window, GtkWidget *text_target)
{
  /* We use the same file selection widget each time, so first
     of all we create it if it hasn't already been created. */

  if (open_filesel_w == NULL)
    open_filesel_w = create_fileselection1 ();

  /* We save a pointer to the main window inside the file selection's
     data list, so we can get it easily in the callbacks. */

  gtk_object_set_data (GTK_OBJECT (open_filesel_w), MainWindowKey, main_window);
  gtk_object_set_data (GTK_OBJECT (open_filesel_w), TargetTextKey, text_target);
  gtk_widget_show (open_filesel_w);
  gdk_window_raise (open_filesel_w->window);
}

void
DisplayTree_helper(struct Tree_node *tr,GtkWidget *parent)
{
  GtkWidget *subtree,*subitem;
  GString *name;
  
  name = g_string_new(NULL);
  g_string_sprintf(name,"%d - ?", tr->id);
  
  /* Create a subtree item, in much the same way */
  subitem = gtk_tree_item_new_with_label (name->str);
  
  /* Connect all GtkItem:: and GtkTreeItem:: signals */
  gtk_signal_connect (GTK_OBJECT(subitem), "select",
		      GTK_SIGNAL_FUNC(on_subitem_select), NULL);
  
  gtk_object_set_data (GTK_OBJECT(subitem), TreeItemKey, tr);
 
  /* Add it to its parent tree */
  gtk_tree_append (GTK_TREE(parent), subitem);

  /* Show it */
  gtk_widget_show (subitem);

  if (tr->child!=NULL)
    {
      subtree = gtk_tree_new();
      gtk_tree_set_selection_mode (GTK_TREE(subtree),
				   GTK_SELECTION_SINGLE);
      gtk_tree_set_view_mode (GTK_TREE(subtree), GTK_TREE_VIEW_ITEM);
      gtk_tree_item_set_subtree (GTK_TREE_ITEM(subitem), subtree);
      DisplayTree_helper(tr->child,subtree);
    }
  
  tr=tr->brother;
  if(tr->id!=-1)
    DisplayTree_helper(tr,parent);
}

void
Raise_tree (GtkWidget *main_window)
{
  GtkWidget *tree;
  GtkWidget *subtree, *item;
  struct Tree_node *tmp;

  tmp=gTree;

  /* We use the same file selection widget each time, so first
     of all we create it if it hasn't already been created. */

  if (tree_w == NULL)
    {
      tree_w = create_tree_w ();

      /* We save a pointer to the main window inside the file selection's
	 data list, so we can get it easily in the callbacks. */

      gtk_object_set_data (GTK_OBJECT (tree_w), MainWindowKey, main_window);

      tree=lookup_widget (GTK_WIDGET (tree_w), "db_tree");

      /* Create a tree item */
      item = gtk_tree_item_new_with_label (gOptions->imm_db_path);
  
      /* Add it to the parent tree */
      gtk_tree_append (GTK_TREE(tree), item);

      /* Show it - this can be done at any time */
      gtk_widget_show (item);

      subtree = gtk_tree_new();
      
      /* This has absolutely no effect, because it is completely ignored 
	 in subtrees */
      gtk_tree_set_selection_mode (GTK_TREE(subtree),
			       GTK_SELECTION_SINGLE);
      
      /* Neither does this, but for a rather different reason - the
	 view_mode and view_line values of a tree are propagated to
	 subtrees when they are mapped.  So, setting it later on would
	 actually have a (somewhat unpredictable) effect */
      gtk_tree_set_view_mode (GTK_TREE(subtree), GTK_TREE_VIEW_ITEM);
      
      /* Set this item's subtree - note that you cannot do this until
	 AFTER the item has been added to its parent tree! */
  
      gtk_tree_item_set_subtree (GTK_TREE_ITEM(item), subtree);
  
      DisplayTree_helper(tmp,subtree);
    }

  gtk_widget_show (tree_w);
  gdk_window_raise (tree_w->window);

}

void
Raise_about_w (GtkWidget *main_window)
{
  /* We use the same file selection widget each time, so first
     of all we create it if it hasn't already been created. */

  if (about_w == NULL)
    about_w = create_about_w ();

  /* We save a pointer to the main window inside the file selection's
     data list, so we can get it easily in the callbacks. */
  gtk_object_set_data (GTK_OBJECT (about_w), MainWindowKey, main_window);

  gtk_widget_show (about_w);
  gdk_window_raise (about_w->window);
}

void
Raise_configure_monitor_w (GtkWidget *main_window)
{
  /* We use the same file selection widget each time, so first
     of all we create it if it hasn't already been created. */

  if (configure_monitor_w == NULL)
    configure_monitor_w = create_configure_monitor ();

  /* We save a pointer to the main window inside the file selection's
     data list, so we can get it easily in the callbacks. */
  gtk_object_set_data (GTK_OBJECT (configure_monitor_w), MainWindowKey, main_window);

  gtk_widget_show (configure_monitor_w);
  gdk_window_raise (configure_monitor_w->window);
}

void
Raise_configure_trace_w (GtkWidget *main_window)
{
  /* We use the same file selection widget each time, so first
     of all we create it if it hasn't already been created. */

  if (configure_trace_w == NULL)
    configure_trace_w = create_configure_trace ();

  /* We save a pointer to the main window inside the file selection's
     data list, so we can get it easily in the callbacks. */
  gtk_object_set_data (GTK_OBJECT (configure_trace_w), MainWindowKey, main_window);

  gtk_widget_show (configure_trace_w);
  gdk_window_raise (configure_trace_w->window);
}

void
Raise_trace_recorder_w (GtkWidget *main_window)
{
  /* We use the same file selection widget each time, so first
     of all we create it if it hasn't already been created. */

  if (trace_recorder_w == NULL)
    trace_recorder_w = create_trace_recorder ();

  /* We save a pointer to the main window inside the file selection's
     data list, so we can get it easily in the callbacks. */
  gtk_object_set_data (GTK_OBJECT (trace_recorder_w), MainWindowKey, main_window);

  gtk_widget_show (trace_recorder_w);
  gdk_window_raise (trace_recorder_w->window);
}
/***************************************************/
/* GLADE : Those functions are registered by GLADE */
/***************************************************/
/***********************************************************/
/* File Menu events.                                       */
/*                                                         */ 
/***********************************************************/

void
on_connect_to_sensor_activate          (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *main_window;
  GtkWidget *statusbar;

  if (!gDeviceOpened || gSensorOpened)
    {
      main_window = lookup_widget (GTK_WIDGET (menuitem), "main_window");
      Raise_connect_to_sensor_w (main_window);
    }
  else
    {
      statusbar = lookup_widget (GTK_WIDGET (menuitem), "statusbar1");
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, device or sensor already opened.");
    }
}

void
on_open_device_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *main_window;
  GtkWidget *statusbar;

  if (!gDeviceOpened || !gSensorOpened)
    {
      main_window = lookup_widget (GTK_WIDGET (menuitem), "main_window");
      Raise_open_device_w (main_window);
    }
  else
    {
      statusbar = lookup_widget (GTK_WIDGET (menuitem), "statusbar1");
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, device already opened.");
    }
}


void
on_close_device_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *statusbar;

  statusbar = lookup_widget (GTK_WIDGET (menuitem), "statusbar1");

  if (!(gDeviceOpened || gSensorOpened))
    {
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, no device or sensor opened.");
      return;
    }
  
  if (gTraceStarted)
    {
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, you have to stop tracing first.");
      return;
    }

   /* If we are here, a device is opened and no trace running*/

  if(CloseDevice()<0)
    gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			"Cannot close device.");
  else
    gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			"Device closed.");
  return;
}


void
on_quit_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  Our_exit(0);
}


void
on_hello_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *statusbar;

  statusbar = lookup_widget (GTK_WIDGET (menuitem), "statusbar1");

  gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, "Text cut to clipboard.");

}


void
on_configure_monitor_activate          (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *main_window;

  main_window = lookup_widget (GTK_WIDGET (menuitem), "main_window");
    
  Raise_configure_monitor_w (main_window);
}


void
on_trace_to_file_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *main_window;
   GtkWidget *statusbar;

  main_window = lookup_widget (GTK_WIDGET (menuitem), "main_window");
  statusbar = lookup_widget (GTK_WIDGET (menuitem), "statusbar1");

  if (!(gDeviceOpened || gSensorOpened))
    {
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, no device opened.");
      return;
    }
  
  if (gTraceStarted)
    {
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, you have to stop tracing first.");
      return;
    }
    
  Raise_configure_trace_w (main_window);

}


void
on_about_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *main_window;
  GtkWidget *statusbar;


      main_window = lookup_widget (GTK_WIDGET (menuitem), "main_window");
      statusbar = lookup_widget (GTK_WIDGET (menuitem), "statusbar1");

      Raise_about_w (main_window);
 
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Check our website for latest release !");
   
}

void
on_learn_profile_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *main_window;

      main_window = lookup_widget (GTK_WIDGET (menuitem), "main_window");
    
      Raise_learn_w (main_window);

}

void
on_start_monitor_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

  GtkWidget *statusbar,*event_log,*mean_box;
  GString   * message;
  statusbar = lookup_widget (GTK_WIDGET (menuitem), "statusbar1");
  event_log = lookup_widget (GTK_WIDGET(statusbar), "text_event");

  if (!(gDeviceOpened || gSensorOpened))
    {
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, no device opened.");
      return;
    }

  if (!gMonitorConfigured)
    {
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, you didn't configure the monitor");
      return;
    }

  
  if (gTraceStarted)
    {
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, trace is already started.");
      return;
    }
  
  /* If we are here, a device is opened and no trace started. */
  /* Let's do it ! */

  if (Start_trace(event_log)<0)
  {
    gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			"Sorry, cannot start the tracing.");
    return;
  }
 
  /* If we are here, the trace started, it's a success !*/
 
  message = g_string_new(NULL);
  g_string_sprintf(message,"INFO: Monitor started.\n");
  
  gtk_text_insert(GTK_TEXT(event_log),
		  NULL,
		  &event_log->style->black,
		  NULL,
		  message->str,-1);

  mean_box = lookup_widget (GTK_WIDGET (menuitem), "mean_value");
  gtk_editable_set_editable(GTK_EDITABLE(mean_box),FALSE);
 
  return;
}

void
on_stop_monitor_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

  GtkWidget *statusbar,*mean_box;
  GtkWidget *event_log;
  GString *message;
  statusbar = lookup_widget (GTK_WIDGET (menuitem), "statusbar1");
  
  if (!gTraceStarted)
    { 
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, no trace running at this time.");
      return;
    }
  
  /* If we are here, a device is opened and a trace running. */

   if(Stop_trace()<0)
    {
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, unable to stop the trace.");
      return;
    }
  
  /* If we are here, the trace stoped, it's a success !*/
  
   event_log = lookup_widget (GTK_WIDGET(statusbar), "text_event");
   message = g_string_new(NULL);
   g_string_sprintf(message,"Monitor stoped.\n");
  
   gtk_text_insert(GTK_TEXT(event_log),
		   NULL,
		   &event_log->style->black,
		   NULL,
		   message->str,-1);

   mean_box = lookup_widget (GTK_WIDGET (menuitem), "mean_value");
   gtk_editable_set_editable(GTK_EDITABLE(mean_box),TRUE);
  
   return;
}


void
on_plot_graph_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *statusbar;
  FILE *out_file;
  char *command;

  statusbar = lookup_widget (GTK_WIDGET (menuitem), "statusbar1");
  
  if (!gOptions->log_to_file)
    { 
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, you were not logging to a file.");
      return;
    }
  
  out_file=fopen("./gnuplot","w");
  fprintf(out_file,"set terminal postscript eps 22\nset size 1.5,1.5\nset yrange [0:1]\nset output '%s.eps'\nplot '%s' w l\nset output\n",
	  gOptions->log_path,
	  gOptions->log_path); 
  fclose(out_file);

  system("gnuplot gnuplot");
  command=(char *)malloc(100);
  sprintf(command,"ghostview %s.eps &",gOptions->log_path);
  system(command);
  free(command); 

}

void
on_show_tree_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *main_window,*statusbar;
  
  statusbar = lookup_widget (GTK_WIDGET (menuitem), "statusbar1");
  
  if (!gMonitorConfigured)
    { 
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, no database loaded at this time - First configure monitor.");
      return;
    }

  main_window = lookup_widget (GTK_WIDGET (menuitem), "main_window");
  
  Raise_tree (main_window);
}


/***********************************************************/
/* Main window events.                                     */
/*                                                         */ 
/***********************************************************/

void
on_mean_value_changed                  (GtkEditable     *editable,
                                        gpointer         user_data)
{
  gint value;

  value= atoi(gtk_editable_get_chars(GTK_EDITABLE(editable),0,-1));
  gOptions->mean_len=value;
}

void
on_an_seq_len_changed                  (GtkEditable     *editable,
                                        gpointer         user_data)
{
  gint value;

  value= atoi(gtk_editable_get_chars(GTK_EDITABLE(editable),0,-1));
  if (value<=gOptions->db_seq_len && value>0)
    gOptions->seq_len=value;
}


gboolean
on_main_window_delete_event            (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  
  Our_exit(0);
  return FALSE;
}


gboolean
on_drawingarea_expose_event            (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{

  if (gSegs==NULL) /* First time we come in here */
    {
      gSegs= NewSegs (500);
    }

  gdk_window_clear (widget->window);
  
  
  gdk_window_set_background (widget->window,
			     &widget->style->black);
  gdk_draw_segments               (widget->window,
				   widget->style->white_gc,
				   gSegs->seg,
				   gSegs->nsegs-1);
  return TRUE;
}


/***********************************************************/
/* Open Device dialog box events.                          */
/*                                                         */ 
/***********************************************************/

gboolean
on_open_device_w_delete_event          (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_widget_hide (gtk_widget_get_toplevel (widget));
  return TRUE;
}


void
on_ok_open_device_clicked              (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *Gtk_Widget;

  gchar     *device_name;
  gchar     *process_name;

  gboolean  execve_on;
  gboolean  fbod_chk;
 
  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "device_name");
  device_name=gtk_editable_get_chars(GTK_EDITABLE (Gtk_Widget),0,-1);

  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "process_name");
  process_name = gtk_editable_get_chars(GTK_EDITABLE (Gtk_Widget),0,-1);

  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "execve_chk");
  execve_on  =  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (Gtk_Widget));

  if (execve_on)
    gOptions->imm_execve_trace=1;
  else
    gOptions->imm_execve_trace=0;

   Gtk_Widget = lookup_widget (GTK_WIDGET (button), "fbod_check");
  fbod_chk  =  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (Gtk_Widget));

  if (fbod_chk)
    {
      gOptions->fbod=1;
    }
  else
    gOptions->fbod=0;

  gOptions->device_name=device_name;
  gOptions->imm_process_name=process_name;

  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (button)));

  Open_device(gtk_widget_get_toplevel (GTK_WIDGET (button)));
}


void
on_cancel_open_device_clicked          (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}


void
on_browse_device_clicked               (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *main_window,*open_device;
  GtkWidget *text_target;

  open_device = gtk_widget_get_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (open_device), MainWindowKey);
  text_target = lookup_widget (GTK_WIDGET (button), "device_name");
  Raise_browse_file (main_window,text_target);
}


void
on_browse_database_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *main_window,*open_device;
  GtkWidget *text_target;

  open_device = gtk_widget_get_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (open_device), MainWindowKey);
  text_target = lookup_widget (GTK_WIDGET (button), "db_path");
  Raise_browse_file (main_window,text_target);
}

void
on_browse_log_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *main_window,*open_device;
  GtkWidget *text_target;

  open_device = gtk_widget_get_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (open_device), MainWindowKey);
  text_target = lookup_widget (GTK_WIDGET (button), "log_path");
  Raise_browse_file (main_window,text_target);

}

void
on_browse_script_clicked               (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *main_window,*open_device;
  GtkWidget *text_target;

  open_device = gtk_widget_get_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (open_device), MainWindowKey);
  text_target = lookup_widget (GTK_WIDGET (button), "script_path");
  Raise_browse_file (main_window,text_target);

}



/********************************************************/
/*                                                      */
/* Browse Device dialog box events.                     */
/********************************************************/

void
on_ok_button1_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *filesel, *main_window,*text_target;
  gchar *filename;

  filesel = gtk_widget_get_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (filesel), MainWindowKey);
  text_target = gtk_object_get_data (GTK_OBJECT (filesel), TargetTextKey);
  gtk_widget_hide (filesel);
  
  filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (filesel));
  gtk_entry_set_text(GTK_ENTRY(text_target),filename);

}


void
on_cancel_button1_clicked              (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (button)));

}


gboolean
on_fileselection1_delete_event         (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (widget)));
  return TRUE;
}

/***********************************************************/
/* Tree events.                                            */
/*                                                         */ 
/***********************************************************/


gboolean
on_tree_w_delete_event                 (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (widget)));
  return TRUE;
}

void
on_subitem_select(GtkWidget *widget,gpointer userdata)
{
  GtkWidget *statusbar;
  struct Tree_node *tmp;
  GString *message;
  
  message = g_string_new(NULL);
 
  statusbar= lookup_widget (GTK_WIDGET (widget), "tree_statusbar");
  tmp = gtk_object_get_data (GTK_OBJECT (widget), TreeItemKey);

  g_string_sprintf(message,"ID: %d Count: %d Proba: %f \n",
		   tmp->id,tmp->count,tmp->proba);

  gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, message->str);
}


/***********************************************************/
/* About Window events                                    */
/*                                                         */ 
/***********************************************************/



gboolean
on_about_w_delete_event                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (widget)));
  return TRUE;
}


void
on_about_ok_button_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
   gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}



/***********************************************************/
/*Learn Window events                                    */
/*                                                         */ 
/***********************************************************/



gboolean
on_learn_w_delete_event                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (widget)));
  return TRUE;
}


void
on_browse_trace_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *main_window,*learn_window;
  GtkWidget *text_target;

  learn_window = gtk_widget_get_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (learn_window), MainWindowKey);
  text_target = lookup_widget (GTK_WIDGET (button), "input_trace");
  Raise_browse_file (main_window,text_target);

}


void
on_browse_output_db_clicked            (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *main_window,*learn_window;
  GtkWidget *text_target;

  learn_window = gtk_widget_get_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (learn_window), MainWindowKey);
  text_target = lookup_widget (GTK_WIDGET (button), "output_db");
  Raise_browse_file (main_window,text_target);

}


void
on_browse_input_db_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *main_window,*learn_window;
  GtkWidget *text_target;

  learn_window = gtk_widget_get_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (learn_window), MainWindowKey);
  text_target = lookup_widget (GTK_WIDGET (button), "input_db");
  Raise_browse_file (main_window,text_target);

}


void
on_ok_learn_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{

  GtkWidget *Gtk_Widget;
  struct Learn_config *config;
  gchar     *input_trace;
  gchar     *input_db;
  gchar     *output_db;
  gchar     *seq_len;
  gboolean  input_db_on;
  gboolean  verbose_on;

  config=(struct Learn_config *)malloc(sizeof(struct Learn_config));
  
  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "input_trace");
  input_trace=gtk_editable_get_chars(GTK_EDITABLE (Gtk_Widget),0,-1);

  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "input_db");
  input_db = gtk_editable_get_chars(GTK_EDITABLE (Gtk_Widget),0,-1);

  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "output_db");
  output_db = gtk_editable_get_chars(GTK_EDITABLE (Gtk_Widget),0,-1);

  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "seq_len");
  seq_len = gtk_editable_get_chars(GTK_EDITABLE (Gtk_Widget),0,-1);
 
  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "input_db_chk");
  input_db_on  =  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (Gtk_Widget));

  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "verbose_chk");
  verbose_on =  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (Gtk_Widget));

  if (verbose_on)
    config->verbose=1;
  else
    config->verbose=0;
  
  if (input_db_on)
    config->load_db=1;
  else
    config->load_db=0;

  config->output_db=1;

  config->trace_path=input_trace;
  config->input_path=input_db;
  config->output_path=output_db;
  config->seq_len=atoi(seq_len);
  config->gtk=1;
  config->gtk_window=(void *)gtk_widget_get_toplevel (GTK_WIDGET (button));

  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (button)));

  do_learn_gtk(config);
}


void
on_cancel_learn_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
 gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}







/***********************************************************/
/*Configure Window events                                    */
/*                                                         */ 
/***********************************************************/




gboolean
on_configure_monitor_delete_event      (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (widget)));
  return TRUE;
}


void
on_ok_configure_monitor_clicked        (GtkButton       *button,
                                        gpointer         user_data)
{


 GtkWidget *Gtk_Widget;

  gchar     *threshold;
  gchar     *db_path;
  gchar     *log_path;
  gboolean  executeonalert;
  gboolean  log_to_file;
  
  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "threshold");
  threshold = gtk_editable_get_chars(GTK_EDITABLE (Gtk_Widget),0,-1);
  
  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "db_path");
  db_path = gtk_editable_get_chars(GTK_EDITABLE (Gtk_Widget),0,-1);
 
  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "executeonalert");
  executeonalert =  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (Gtk_Widget));

  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "log_chk");
  log_to_file  =  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (Gtk_Widget));
  
  if (log_to_file)
    {
      gOptions->log_to_file=1;
      Gtk_Widget = lookup_widget (GTK_WIDGET (button), "log_path");
      log_path  = gtk_editable_get_chars(GTK_EDITABLE (Gtk_Widget),0,-1);
      gOptions->log_path=log_path;
    }
  else
    gOptions->log_to_file=0;


    if (executeonalert)
    {
      gOptions->exec_script=1;
      Gtk_Widget = lookup_widget (GTK_WIDGET (button), "script_path");
      gOptions->script_path= gtk_editable_get_chars(GTK_EDITABLE (Gtk_Widget),0,-1);
    }
  else
    gOptions->exec_script=0;

  gOptions->imm_threshold=(atof (threshold));
  gOptions->imm_db_path=db_path;

  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (button)));

  Configure_monitor(gtk_widget_get_toplevel (GTK_WIDGET (button)));

}


void
on_cancel_configure_monitor_clicked    (GtkButton       *button,
                                        gpointer         user_data)
{
 gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}


/***********************************************************/
/*Configure Tracer events                                  */
/*                                                         */ 
/***********************************************************/


gboolean
on_configure_trace_delete_event        (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (widget)));
  return TRUE;
}


void
on_ok_configure_trace_clicked          (GtkButton       *button,
                                        gpointer         user_data)
{

  GtkWidget *Gtk_Widget;
  GtkWidget *main_window;
  GtkWidget *configure_tracer;
  GtkWidget *statusbar;
  gchar     *trace_destination;
  gchar     *max_events;
  gboolean  max_events_chk;


  configure_tracer=gtk_widget_get_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (configure_tracer), MainWindowKey);
  statusbar = lookup_widget (GTK_WIDGET (main_window), "statusbar1");


  if (gTracerConfig==NULL)
    gTracerConfig=(tracer_options*)malloc(sizeof(tracer_options));

  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "maxevents");
  max_events = gtk_editable_get_chars(GTK_EDITABLE (Gtk_Widget),0,-1);
  
  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "trace_destination");
  trace_destination = gtk_editable_get_chars(GTK_EDITABLE (Gtk_Widget),0,-1);
 
  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "maxevents_chk");
  max_events_chk  =  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (Gtk_Widget));

  if (max_events_chk)
    gTracerConfig->set_max_events=TRUE;
  else
     gTracerConfig->set_max_events=FALSE;


  gTracerConfig->log_path=trace_destination;
 
  gTracerConfig->max_events=(atof (max_events));


  gBuffer=(struct event *)malloc(gOptions->buffer_size*sizeof(struct event));

  /* The device as been reset (syscall_count to 0), we reset the last alert.*/
  gLastAlert=0;

  /* Open and set up the log file if the user asked to log. */
 
  if ((gLogFile= fopen(gTracerConfig->log_path,"w"))==NULL)
      {
	gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "ERROR: Cannot open log file. !");
	return;
      }
  
  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (button)));
   
  Raise_trace_recorder_w (main_window);
}


void
on_cancel_configure_trace_clicked      (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (button)));
  
}


void
on_browse_trace_destination_clicked    (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *main_window,*tracer_window;
  GtkWidget *text_target;

  tracer_window = gtk_widget_get_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (tracer_window), MainWindowKey);
  text_target = lookup_widget (GTK_WIDGET (button), "trace_destination");
  Raise_browse_file (main_window,text_target);

}

/***********************************************************/
/*Trace recorder events                                    */
/*                                                         */ 
/***********************************************************/


gboolean
on_trace_recorder_delete_event         (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (widget)));
  return TRUE;
}


void
on_start_trace_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *statusbar,*event_log,*main_window,*tracer_window,*label;
  GString   * message;


  tracer_window = gtk_widget_get_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (tracer_window), MainWindowKey);
  label=lookup_widget (GTK_WIDGET (tracer_window), "trace_calls_label");
  statusbar = lookup_widget (GTK_WIDGET (main_window), "statusbar1");
  event_log = lookup_widget (GTK_WIDGET(statusbar), "text_event");

  if (!(gDeviceOpened || gSensorOpened))
    {
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, no device opened.");
      return;
    }
  
   if (gTracerConfig==NULL)
    {
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, tracer not configured.");
      return;
    }

  if (gTraceStarted)
    {
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, trace is already started.");
      return;
    }
  
  /* If we are here, a device is opened and no trace started. */
  /* Let's do it ! */

  if (Start_trace_record(label)<0)
  {
    gtk_text_insert(GTK_TEXT(event_log),
		  NULL,
		  &event_log->style->black,
		  NULL,
		  "Sorry, cannot start the tracing.",-1);
    return;
  }
 
   gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			"Tracing started");

  /* If we are here, the trace started, it's a success !*/
 
  message = g_string_new(NULL);
  g_string_sprintf(message,"INFO: Recording of trace started.\n");
  
  gtk_text_insert(GTK_TEXT(event_log),
		  NULL,
		  &event_log->style->black,
		  NULL,
		  message->str,-1);
}


void
on_stop_trace_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{

  GtkWidget *statusbar,*tracer_window,*main_window;
  GtkWidget *event_log;
  GString *message;


  tracer_window = gtk_widget_get_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (tracer_window), MainWindowKey);
  statusbar = lookup_widget (GTK_WIDGET (main_window), "statusbar1");
  event_log = lookup_widget (GTK_WIDGET(statusbar), "text_event");
  
  if (!gTraceStarted)
    { 
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, no trace running at this time.");
      return;
    }
  
  /* If we are here, a device is opened and a trace running. */

   if(Stop_trace_record()<0)
    {
      gtk_statusbar_push (GTK_STATUSBAR (statusbar), 1, 
			  "Sorry, unable to stop the trace.");
      return;
    }
  
  /* If we are here, the trace stoped, it's a success !*/
  
   event_log = lookup_widget (GTK_WIDGET(statusbar), "text_event");
   message = g_string_new(NULL);
   g_string_sprintf(message,"INFO: Recording of trace stoped.\n");
  
   gtk_text_insert(GTK_TEXT(event_log),
		   NULL,
		   &event_log->style->black,
		   NULL,
		   message->str,-1);
  
   gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (button)));
   
   return;
  
}


/***********************************************************/
/*Configure sensor events                                  */
/*                                                         */ 
/***********************************************************/


gboolean
on_open_sensor_delete_event            (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (widget)));
  return TRUE;
  
}


void
on_ok_open_sensor_clicked              (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *Gtk_Widget;

  gchar     *sensor_name;

  Gtk_Widget = lookup_widget (GTK_WIDGET (button), "sensor_name");
  sensor_name=gtk_editable_get_chars(GTK_EDITABLE (Gtk_Widget),0,-1);

  gOptions->sensor_name=sensor_name;
 

  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (button)));

  Open_sensor(gtk_widget_get_toplevel (GTK_WIDGET (button)));
}


void
on_cancel_open_sensor_clicked          (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_hide (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}


void
on_browse_sensor_clicked               (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *main_window,*open_sensor;
  GtkWidget *text_target;

  open_sensor = gtk_widget_get_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (open_sensor), MainWindowKey);
  text_target = lookup_widget (GTK_WIDGET (button), "sensor_name");
  Raise_browse_file (main_window,text_target);
}

