#include "common.h"

struct Mean *NewMean (int nb_el)
{
  struct Mean *tmp;
  
  tmp=(struct Mean *)malloc(sizeof(struct Mean));
  tmp->data=(float *)malloc(nb_el*sizeof(float));
  tmp->mean=0.0;
  tmp->total=nb_el;
  tmp->actual=0;
  tmp->first=0;
  tmp->minus=1.0;
  
  return tmp;
}

void UpdateMean (struct Mean *mean,float value)
{ 
  float old;
  float tmp;

  if (mean->actual==mean->total)
    {
      old=*(mean->data+mean->first);
      tmp=((mean->mean*mean->total)-old+value)/mean->total;
      mean->mean=tmp;
      *(mean->data+mean->first)=value;
      mean->first+=1;

      if (mean->first==mean->total)
	mean->first=0;

      if (mean->minus>mean->mean) 
	mean->minus=mean->mean;
    }
  else
    {
      mean->mean=((mean->mean * mean->actual)+value)/(mean->actual+1);
      *(mean->data+mean->actual)=value;
      mean->actual+=1;
      mean->minus=mean->mean;
    }

}

/********************************************************************/
/* int * AppendToSequence(struct Seq_stream *as, int syscall)       */
/*                                                                  */
/* Ajoute l'appel systeme pour creer une sequence.                  */ 
/* Si syscall=-1, reinitialise la sequence.                         */
/* Si une sequence complete existe, renvoi un pointeur vers celle-ci*/
/********************************************************************/

int * AppendToSequence(struct Seq_stream *as, int syscall,int seq_len)
{
  int i;

  if (syscall == -1)
    as->in_seq=0;
  else
    if (as->in_seq< seq_len)
      as->seq[as->in_seq++]=syscall;
    else
      {
	for (i=0;i<seq_len-1;i++)
	  as->seq[i]=as->seq[i+1];
	as->seq[seq_len-1]=syscall;
      }
  if (as->in_seq==seq_len)
    return(as->seq);
  else
    return NULL;
}

/********************************************************************/
/* (int *) GetActiveSeq(int pid, int syscall, struct Seq_stream as,int seq_len) */
/*                                                                  */
/* Cette fonction retourne un pointeur vers la sequence actuelle    */ 
/* si celle-ci est complete.                                        */
/* Cree une nouvelle stream si necessaire.                          */
/********************************************************************/

int * GetActiveSeq(int pid, int syscall, struct Seq_stream *as,int seq_len)
{
  struct Seq_stream *tmp_stream;
  while(as->pid!=-1)
    {
      if (as->pid == pid)
	return(AppendToSequence(as,syscall,seq_len));
      else
	as=as->next;
    }
  
  //No stream for this PID, we have to create a new one.

  tmp_stream= (struct Seq_stream *) malloc(sizeof(struct Seq_stream));
  tmp_stream->next=NULL;
  tmp_stream->pid=-1;
  tmp_stream->seq=(int *)malloc(seq_len*sizeof(int));
  as->next=tmp_stream;
  as->pid=pid;
  as->in_seq=0;

  return (int *) AppendToSequence(as,syscall,seq_len);

}

void UpdateProbas_Node(struct Tree_node *tree, int total)
{
  tree->proba=((float) tree->count / (float) total);
  if (tree->brother!=NULL)
    UpdateProbas_Node(tree->brother, total);
  if (tree->child!=NULL)
    {
      if (PROBA_LOCAL) 
	UpdateProbas_Node(tree->child, tree->count);
      else
	UpdateProbas_Node(tree->child, total);
    }
}

void UpdateProbas(struct Tree_node *tree)
{
  int total_seq=0;
  struct Tree_node *tmp;
  tmp=tree;
  
  while (tmp->id!=-1)
    {
      total_seq+=tmp->count;
      tmp=tmp->brother;
    }
 
  UpdateProbas_Node(tree,total_seq);  
}


void ProcessProba(struct Tree_node *tree,int *seq,float *proba,int seq_len)
{
  int j=0;
  struct Tree_node *tmp;
  float value=0;

  if ((tmp=GetNode(tree,seq[0]))==NULL)
    value=0;
  else
    for(j=1;j<seq_len;j++)
      {
	if((tmp=GetNode(tmp->child,seq[j]))==NULL)
	  {value=0;break;}
	else
	  value=tmp->proba;
      }
  
  *proba=value;

}


int LoadTreeNode(struct Tree_node *tree, FILE *in, int level,int seq_len)
{
  int up_to;
  int arg1,arg2;

  if (level<seq_len-1)
    {
      fscanf(in,"[%d %d] ",&arg1,&arg2);
      tree->id=arg1;
      tree->count=arg2;
      tree->child=NewTreeNode();
      tree->brother=NewTreeNode();
      up_to=LoadTreeNode(tree->child, in, level+1,seq_len);
    }
  else
    {
      fscanf(in,"[%d %d] ",&arg1,&arg2);
      tree->id=arg1;
      tree->count=arg2;
      tree->child=NULL;
      tree->brother=NewTreeNode();
      fscanf(in,"%d ",&arg1);
      if (arg1==-1)
	{fscanf(in,"\n");
	up_to=0;
	}
      else
	up_to=seq_len+arg1+1;
	
    }

  while (up_to==level && up_to!=0)
    up_to=LoadTreeNode(tree->brother, in, level,seq_len);
      
  return(up_to);
}

int LoadTree(struct Tree_node *tree, char *input_path)
{
  struct Tree_node *tmp;
  FILE *in_file;
  int arg1;
  tmp=tree;
 
  if ((in_file = fopen(input_path,"r")) ==NULL)
    return(-1);

  fscanf(in_file,"#DBrev: %d\n",&arg1);
  if (arg1!=1)
    {
      fprintf(stderr,"Invalid database version %d \n",arg1);
      return(-2);
    }
 
  fscanf(in_file,"#DBseq_len: %d\n",&arg1);
  gOptions->seq_len=arg1;

  fscanf(in_file,"%d\n",&arg1);
  while (arg1!=-1)
    {
      tmp->brother=NewTreeNode();
      LoadTreeNode(tmp,in_file,0,gOptions->seq_len);
      tmp=tmp->brother;
      fscanf(in_file,"%d\n",&arg1);
    }

  if (fclose(in_file)==EOF) 
    return(-1);
  
  return(0);
}

struct Tree_node * NewTreeNode()
{
  struct Tree_node *tmp;
  
  tmp= (struct Tree_node *) malloc(sizeof(struct Tree_node));
  tmp->id=-1;
  tmp->proba=0;
  tmp->count=0;
  tmp->child=NULL;
  tmp->brother=NULL;
  
  return tmp;
}

struct Tree_node *GetNode(struct Tree_node *tree,int id)
{
  while(tree->id!=-1)
    if (tree->id==id)
      return(tree);
    else
      tree=tree->brother;
  
  return(NULL);
}
