#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <gtk/gtk.h>

#define SAMPLE_TIME 3
#define PROBA_LOCAL 1
#define GNUPLOT "./gnuplot"

#define BUFFER_LENGTH 65540

#define IMM_SAY_HELLO 1
#define IMM_SAY_HI    2
#define IMM_BUFFER_RESET 3
#define IMM_START 4
#define IMM_STOP 5
#define IMM_SET_NAME 6
#define IMM_EXECVE_TRACE_ON 7
#define IMM_EXECVE_TRACE_OFF 8
#define IMM_BUFFER_SIZE 9

#define LOG_EVENT(msg)     {gtk_text_insert(GTK_TEXT(event_log),NULL, &event_log->style->black, NULL, msg,-1);error=1;}

struct event {
  int count;
  int pid;
  int call;
  unsigned long eip;    
};

typedef struct _options
{
  char *device_name;
  char *sensor_name;
  int seq_len;
  int db_seq_len;
  int mean_len;
  int log_to_file;
  
  char *imm_process_name;
  int imm_execve_trace;
  float imm_threshold;
  int buffer_size;

  int exec_script;
  char *script_path;

  guint32 imm_callback;
  char *log_path;
  char *imm_db_path;

  int fbod;
  unsigned long fbod_threshold_up;
  unsigned long fbod_threshold_down;
} options;

typedef struct _sensor
{
  struct sockaddr saddr;
  int s;
  FILE *in;
} sensor;

typedef struct _tracer
{
  char *log_path;
  int max_events;
  gboolean set_max_events;
  gboolean tracing;
} tracer_options;

struct Seq_stream {
  int pid;
  int *seq;
  int in_seq;
  struct Seq_stream *next;
};

struct Tree_node {
  int id;
  int count;
  float proba;
  int seq_id;
  struct Tree_node *brother;
  struct Tree_node *child;
};  

struct Mean {
  float * data;
  float   mean;
  float   minus;
  int     total;
  int     actual;
  int     first;
};

struct Segments {
  GdkSegment *seg;
  gint  nsegs;
};


/* Global variables */

struct Tree_node *gTree;
int      gImmDevice;       /* File descriptor to tracing driver */
struct event *gBuffer;
options* gOptions;       /* The user given options */
sensor* gSensor;
tracer_options* gTracerConfig;

struct Seq_stream gActive_stream;
time_t   gStartTime;
FILE * gLogFile;
FILE * gSystemStatus;

struct Mean *gMean;
struct Segments *gSegs;
GdkColormap *gCmap;
GdkColor gColorRed;
GdkColor gColorGreen;
int gLastAlert;

/* Fonctions prototypes */

struct Tree_node * NewTreeNode(void);
struct Tree_node * GetNode(struct Tree_node * ,int);
struct Mean * NewMean (int);
void UpdateMean (struct Mean *,float );
int * AppendToSequence(struct Seq_stream *, int ,int );
int * GetActiveSeq(int, int , struct Seq_stream *,int);
void UpdateProbas_Node(struct Tree_node *, int );
void UpdateProbas(struct Tree_node *);
void ProcessProba(struct Tree_node *,int *,float *,int );
int LoadTreeNode(struct Tree_node *, FILE *, int,int );
int LoadTree(struct Tree_node *, char *);
struct Tree_node * NewTreeNode();

