/*  Ims-trace.c - Command line tool to record a trace.

    ImSafe - Host Based Intrusion Detection
    Version 0.2.2 - Alpha

    Copyright (C) 2001  Laurent Eschenauer 
    <laurent_e@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#define VERSION "0.2.2 Alpha"

/* Necessary headers */
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>


#define TRUE 1
#define FALSE 0

/* Temps en secondes entre deux lectures du device.*/
#define SAMPLE_TIME 3

#define BUFFER_LENGTH 4098 

#define IMM_SAY_HELLO 1
#define IMM_SAY_HI    2
#define IMM_BUFFER_RESET 3
#define IMM_START 4
#define IMM_STOP 5
#define IMM_SET_NAME 6
#define IMM_EXECVE_TRACE_ON 7
#define IMM_EXECVE_TRACE_OFF 8
#define IMM_BUFFER_SIZE 9

struct event {
  int count;
  int pid;
  int call;
  unsigned long eip;
};

typedef struct _options
{
  int run_as_deamon;
  char *device_name;
  
  char *imm_process_name;
  int imm_execve_trace;
  int buffer_size;

  char *log_path;

} options;

/* Global variables */
int      gPID;           /* The daemon's PID */
int      gImmDevice;       /* File descriptor to tracing driver */
struct event *gBuffer;
FILE     *gTracDataFile;  /* The file to which tracing data is dumped */

options* gOptions;       /* The user given options */


options* CreateOptions(void)
{
  options* pOptions;    /* Options to be created */

  /* Allocate space for new options */
  if((pOptions = (options*) malloc(sizeof(options))) == NULL)
    {
    printf("Unable to allocate space for options \n");
    exit(1);
    }

  /* Initialize options */
  pOptions->run_as_deamon         = FALSE;
  pOptions->device_name           = "/dev/immsec";

  pOptions->imm_process_name      = NULL;
  pOptions->imm_execve_trace      = FALSE;
  pOptions->buffer_size           = BUFFER_LENGTH;
  
  pOptions->log_path              = NULL;

  /* Give the options to the caller */
  return pOptions;
}

void ParseOptions(int pmArgc, char** pmArgv, options* pmOptions)
{
  int     i;      /* Generic indexes */
  
  i=1;

  if (pmArgc<2)
      {
          printf("Usage : %s -i <device> -o <out_path> -d[deamon] -e <0,1> -p <process_name>\n",pmArgv[0]);
	  exit(1);
    }
    
  while (i<pmArgc)
    {
      switch (*(pmArgv[i]+1))
	{
	case 'd':
	  pmOptions->run_as_deamon=TRUE;
	  i++;
	  break;
	case 'o':
	  pmOptions->log_path=pmArgv[i+1];
	  i+=2;
	  break;
	case 'i':
	  pmOptions->device_name=pmArgv[i+1];
	  i+=2;
	  break;
	case 'e':
	  if (atoi(pmArgv[i+1])==0)
	    pmOptions->imm_execve_trace=FALSE;
	  else
	    pmOptions->imm_execve_trace=TRUE;
	  i+=2;
	  break;
	case 'p' :
	  pmOptions->imm_process_name=pmArgv[i+1];
	  i+=2;
	  break;
	default: 
	  printf("Unknow option : -%c \n Usage : %s -i device -o out_path -d deamon -e [0,1] -p process_name\n",pmArgv[i][1],pmArgv[0]);
	  exit(1);
	};
    };
}

void ApplyConfig(int pmFileID, options* pmOptions)
{
  

   if (pmOptions->imm_process_name!=NULL)
     {
       if(ioctl(pmFileID, IMM_SET_NAME, pmOptions->imm_process_name))
       {
	 /* Don't go any further */
	 printf("ImTRACER: Unable to set process name to trace. \n");
	 exit(1);
       }
     }
   else
     {
       /* Don't go any further */
       printf("ImTRACER: Invalid process name %s. \n",pmOptions->imm_process_name);
       exit(1);
     }

   if (pmOptions->imm_execve_trace)
     {
       if(ioctl(pmFileID, IMM_EXECVE_TRACE_ON, 0))
	 {
	   /* Don't go any further */
	   printf("ImTRACER: Unable to set execve_on. \n");
	   exit(1);
	 }
     }
   
   else
     if(ioctl(pmFileID, IMM_EXECVE_TRACE_OFF, 0))
       {
	 /* Don't go any further */
	 printf("ImTRACER: Unable to set execve_off. \n");
	 exit(1);
       }
   
}   
void DumpBuffer(void)
{

  int SizeRead,i;
  int syscall_count,pid,syscall;
  unsigned long eip,esp;

  SizeRead = read(gImmDevice, gBuffer, gOptions->buffer_size*sizeof(struct event));
  ioctl(gImmDevice,IMM_BUFFER_RESET,0);

  if (SizeRead >1)
    {
      i=0;
      while (i*sizeof(struct event)<SizeRead-1*sizeof(struct event))
	{
	  syscall_count=gBuffer[i].count;
	  pid=gBuffer[i].pid;
	  syscall=gBuffer[i].call;

	  fprintf(gTracDataFile,"%d %d %d\n",syscall_count,pid, syscall);

	  i+=1;
	};
    }
}
   

void SigHandler(int pmSignalID)
{
  int         lSizeRead;               /* Quantity of data read from driver */
  sigset_t    lSigSet;                 /* Set of signals handled */

  printf("Received signal %d\n",pmSignalID);
  /* Depending on the received signal */
  switch(pmSignalID)
    {
    case SIGIO :
      /* Read from the tracing driver to the data file */
      printf("New data in buffer, cool !\n");
      ioctl(gImmDevice, IMM_BUFFER_RESET, 0);
      break;

    case SIGINT :
    case SIGKILL :
    case SIGQUIT :
    case SIGTERM :
    case SIGALRM :
      /* Stop the driver */
      ioctl(gImmDevice, IMM_STOP, 0);
      ioctl(gImmDevice, IMM_BUFFER_RESET, 0);
      
      /* Dump what is left */
      DumpBuffer();
      /* OK we're out of here */
      close(gImmDevice);
      fclose(gTracDataFile);

      /* Tell the user that we're done */
      fprintf(stdout, "\nImTRACER: End of tracing \n");
      _exit(0);
      break;
      
    default :
      fprintf(stdout, "Received unknown signal \n");
    }
}



int main(int pmArgc, char** pmArgv)
{
  struct sigaction lSigAct;    /* Structure passed on to sigaction() */
  FILE *stream;
  pid_t lPID;

  /* Set the stdout buffer to NULL to get message immediately */
  setbuf(stdout,NULL);

  /* Allocate space for the options */
  gOptions = CreateOptions();

  /* Parse the given options */
  ParseOptions(pmArgc, pmArgv, gOptions);

  fprintf(stdout,"\n ImTRACER version %s \n",VERSION);

  if (gOptions->run_as_deamon)
    if((lPID = fork()) < 0)
      {
	/* Print an error message */
	printf("ImTRACER: Unable to become daemon \n");
	/* Exit */
	exit(1);
      }
    else if(lPID)
      {
	/* This is the parent, Die */
	fprintf(stdout,"ImTRACER: Running as deamon.\n");
	fprintf(stdout,"Kill - 3 PID to terminate the process\n");
	exit(0);
      }
  
  /* Become session leader */
  gPID = setsid();

  /* Initialize structure passed to sigaction() */
  lSigAct.sa_handler = SigHandler;          /* Signal hanlder */
  sigfillset(&(lSigAct.sa_mask));           /* Signal mask */
  lSigAct.sa_flags = 0;                     /* Action flags (No re-register or unblock) */

  /* Set up the signal handler */
  if(sigaction(SIGIO, &lSigAct, NULL)) exit(1);
  if(sigaction(SIGTERM, &lSigAct, NULL)) exit(1);
  if(sigaction(SIGALRM, &lSigAct, NULL)) exit(1);
  if(sigaction(SIGQUIT, &lSigAct, NULL)) exit(1);
  if(sigaction(SIGINT, &lSigAct, NULL)) exit(1);

  
  /* Open the tracing driver */
  if((gImmDevice = open(gOptions->device_name, O_RDONLY, 0)) < 0)
    {
    /* Device ain't good */
    printf("\nUnable to open specified device : %s \n", gOptions->device_name);
    exit(1);
    }

  /* We opened the driver */
  printf("ImTRACER: Trace driver open \n");

  /* Open the output file*/
  if((gTracDataFile = fopen(gOptions->log_path, "a")) ==NULL)
    {
    /* Device ain't good */
    printf("ImTRACER: Unable to open output file : %s \n", gOptions->log_path);
    exit(1);
    }

  /* Apply the configuration to the driver. */
  ApplyConfig(gImmDevice, gOptions);

  /* Create place to dump the data */
  gBuffer=(struct event *)malloc(gOptions->buffer_size*sizeof(struct event));

  /* Fire up the tracing */
  if(ioctl(gImmDevice, IMM_START, 0) < 0)
    {
    /* Don't go any further */
    printf("ImTRACER: Unable to start the tracing \n");
    exit(1);
    }

  /* Sleep forever */
  while(1)
    {
      DumpBuffer();
      sleep(SAMPLE_TIME);
    }
 
  /* SHOULD NEVER GET HERE! */
  exit(1);
}

