/*  learning.c

    ImSafe - Host Based Intrusion Detection
    Version 0.2.2 - Alpha

    Copyright (C) 2001  Laurent Eschenauer 
    <laurent_e@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#define SEQ_LEN 6

/* Si 0 alors il y a calcul des frequences absolues */
#define PROBA_LOCAL 1


#include <stdio.h>                 /*Pour les entrees et sorties*/
#include <sys/types.h>             /*Types systemes:pid,..      */
#include <math.h>
#include <time.h>
#include <string.h>
#include <gtk/gtk.h>
#include "learning.h"
#include "common.h"
#include "support.h"

static int  verbose=0; /* Verbose mode*/
static int  seq_len=SEQ_LEN;
extern const gchar *MainWindowKey;


static void SaveTree(struct Tree_node *,char *);
static void SaveTreeHelper(struct Tree_node *, int,FILE * );
static int AddToTree(struct Tree_node *, int *);


/*************************************************
 *
 *
 * Fonctions du module d'apprentissage
 *
 *
 ************************************************/

int do_learn_gtk(struct Learn_config *config)
{

  /* ============= variables locales  ====================*/ 
  GtkWidget *main_window,*event_log;
  GString *message;

  int i,pid,syscall,syscall_count,new_seq;
  int ord=0;
  int total=0;
  int *current_seq;
  int *seq_buff;

  FILE *trace_file;

  struct Seq_stream active_stream;
  struct Tree_node *tree;
 
  tree=NewTreeNode();
  setbuf(stdout,NULL);

  verbose=config->verbose;
  seq_len=config->seq_len;



  main_window = gtk_object_get_data (GTK_OBJECT ((GtkWidget *)config->gtk_window), MainWindowKey);
  event_log = lookup_widget (main_window, "text_event");
 
  gtk_text_insert(GTK_TEXT(event_log),
		  NULL, 
		  &event_log->style->black, 
		  NULL,
		  "LEARNING: Starting learning profile\n",-1);

  /* A buffer for a sequence of length seq_len */
  seq_buff=(int *) malloc (seq_len*sizeof(int));
  message = g_string_new(NULL);

  /* We prepare a stream vector (we'll need it to learn AND analyse) */
  /* This is linked list contains sequences to analyse for each PID  */
  /* in_seq is the number of calls already in that sequence.         */

  active_stream.pid=-1;
  active_stream.next=NULL;
  active_stream.in_seq=0;
  active_stream.seq=(int *) malloc(seq_len*sizeof(int));
  
  /* Loading the database and updating the proba at leaf nodes.*/
  if (config->load_db)
    {
      LoadTree(tree,config->input_path);
      UpdateProbas(tree);
    }


  if ((trace_file=fopen(config->trace_path,"r")) ==NULL)
    exit(errorHandler(2,0,config->trace_path));

  /* We have to lean a new profile */

  ord=0;total=0;
  while (fscanf(trace_file,"%d %d %d",&syscall_count,&pid,&syscall) !=EOF)
    { 
      total+=1;
      current_seq=GetActiveSeq(pid,syscall,&active_stream,seq_len);
      if (current_seq!=NULL) //We have a seq ready.
	{ 
	  new_seq=AddToTree(tree,current_seq);
	  if (config->verbose && new_seq)
	    {
	      printf("We added: ");    

	      for(i=0;i<seq_len;i++)
		  printf("%d ",current_seq[i]);

	      printf("\n"); 

	    }
	}
    }
    
    if (config->output_db )
      SaveTree(tree,config->output_path);

    gtk_text_insert(GTK_TEXT(event_log),
			      NULL, 
			      &event_log->style->black, 
			      NULL,
			      "LEARNING: New database saved to file",-1); 

    free(active_stream.seq);
    free(seq_buff);

    return(0);
}

int do_learn(struct Learn_config *config)
{

  /* ============= variables locales  ====================*/ 

  int i,pid,syscall,new_seq,syscall_count;
  int ord=0;
  int total=0;
  int *current_seq;
  int *seq_buff;

  FILE *trace_file;

  struct Seq_stream active_stream;
  struct Tree_node *tree;

  tree=NewTreeNode();
  setbuf(stdout,NULL);

  verbose=config->verbose;
  seq_len=config->seq_len;

  /* A buffer for a sequence of length seq_len */
  seq_buff=(int *) malloc (seq_len*sizeof(int));


  /* We prepare a stream vector (we'll need it to learn AND analyse) */
  /* This is linked list contains sequences to analyse for each PID  */
  /* in_seq is the number of calls already in that sequence.         */

  active_stream.pid=-1;
  active_stream.next=NULL;
  active_stream.in_seq=0;
  active_stream.seq=(int *) malloc(seq_len*sizeof(int));
  
  /* Loading the database and updating the proba at leaf nodes.*/
  if (config->load_db)
    {
      LoadTree(tree,config->input_path);
      UpdateProbas(tree);
    }


  if ((trace_file=fopen(config->trace_path,"r")) ==NULL)
    exit(errorHandler(2,0,config->trace_path));

  /* We have to lean a new profile */

  ord=0;total=0;
  while (fscanf(trace_file,"%d %d %d",&syscall_count,&pid,&syscall) !=EOF)
    { 
      total+=1;
      current_seq=GetActiveSeq(pid,syscall,&active_stream,seq_len);
      if (current_seq!=NULL) //We have a seq ready.
	{ 
	  new_seq=AddToTree(tree,current_seq);
	  if (config->verbose && new_seq)
	    {
	      printf("We added : ");
	      for(i=0;i<seq_len;i++)
		printf("%d ",current_seq[i]);
	      printf("\n");
	    }
	}
    }
    
    if (config->output_db )
      SaveTree(tree,config->output_path);

    free(active_stream.seq);
    free(seq_buff);

    return(0);
}



/********************************************************************/
/* int AddToTree(struct Tree_node *tr, int * seq)                    */
/*                                                                   */
/* Ajoute la sequence pointee par seq dans l'arbre Tree_Node.        */

/* Retourne 0 si la seq etait deja presente.                         */
/* Retourne 1 si de nouvelles branches ont ete ajoutees.             */
/* Dans les deux cas, chaque feuille du chemin suivit voit son count */
/* incremente de 1.                                                  */
/********************************************************************/

static int AddToTree(struct Tree_node *tr, int * seq)
{
  int i=0;
  int new_seq=0;
  struct Tree_node *new;
  struct Tree_node *first_of_line;

  first_of_line=tr;
  
  while(i<seq_len)
    {
      if (tr->id == seq[i])
	  {
	    tr->count+=1;
	    tr=tr->child;
	    first_of_line=tr;
	    i++;
	  }
      else if (tr->id == -1)
	{
	  tr=first_of_line;
	  while (tr->id < seq[i] && tr->id !=-1)
	    {
	      tr=tr->brother;
	    }

	  new=NewTreeNode();
	  *new=*tr;
	  tr->id=seq[i];
	  tr->count=1;
	  tr->brother=new;
		 
	  if (i==seq_len-1)
	      tr->child=NULL;
	  else
	    tr->child=NewTreeNode();

	  tr=tr->child;
          first_of_line=tr;
	  new_seq=1;
	  i++;
	}
      else
	tr=tr->brother;
    }
 
  return(new_seq);

}


static void SaveTree(struct Tree_node *tree, char *output_path)
{
  struct Tree_node *tmp;
  FILE *out_file;
  tmp=tree;
 
  
  if ((out_file = fopen(output_path,"w")) ==NULL)
    exit(errorHandler(2,0,output_path));

  fprintf(out_file,"#DBrev: 1\n");
  fprintf(out_file,"#DBseq_len: %d\n",seq_len);
  while(tmp->id!=-1)
    {
      fprintf(out_file,"%d \n[%d %d] ",tmp->id,tmp->id,tmp->count);
      SaveTreeHelper(tmp->child,1,out_file);
      fprintf(out_file,"-1 \n");
      tmp=tmp->brother;
    }
  fprintf(out_file,"-1\n"); //Plus de noeud root dans la base de donnee.
  
  if (fclose(out_file)==EOF) 
    exit(errorHandler(3,0,output_path));
}

static void SaveTreeHelper(struct Tree_node *tree, int level,FILE *out_file)
{
  fprintf(out_file,"[%d %d] ",tree->id,tree->count);
  if (level<seq_len-1)
    SaveTreeHelper(tree->child, level+1,out_file);

  tree=tree->brother;
  if(tree->id!=-1)
    {
      fprintf(out_file,"-%d ",seq_len - level +1 );
      SaveTreeHelper(tree,level,out_file);
 
    }
}
 

/********************************************************************/
/* void errrorhandler(int type,long arg, char * argString)          */
/*                                                                  */
/********************************************************************/

int errorHandler(int type, long arg, char * argString)
{
  switch (type)
    {
    case 1:
      fprintf(stderr,"usage %s [-l(earn) seq_len] [-a mean seq_len] -v(erbose) -s(tats) -g(raph) [-i in_db] [-o out_db] \n",argString);
      return(1);
    case 2:
      fprintf(stderr,"error opening file %s\n",argString);
      return(2);
    case 3:
      fprintf(stderr,"error closing %s.\n",argString);
      return(3);
    case 4:
      fprintf(stderr,"Erreur allocation shm pour %s de taille %d\n",argString, (int)arg);
      return(4);
    case 5:
      fprintf(stderr,"Erreur d'attach shm pour %s.\n",argString);
      return(5);
     case 6:
      fprintf(stderr,"Erreur pour clear shmid de ID=%d.\n", (int)arg);
      return(6); 
    case 7:
      fprintf(stderr,"Erreur lors d'un fork, on a cree %d %s\n",(int)arg,argString);
      return(7);
    case 8:
      fprintf(stderr,"An error occured in a child, deadlock risk, quitting...\n");
      return(8);
   };
  return(0);
}
