/*  main.c

    ImSafe - Host Based Intrusion Detection
    Version 0.2.2 - Alpha

    Copyright (C) 2001  Laurent Eschenauer 
    <laurent_e@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/ioctl.h>

#include <gtk/gtk.h>

#include "interface.h"
#include "support.h"
#include "common.h"



options* CreateOptions(void)
{
  options* pOptions;    /* Options to be created */

  /* Allocate space for new options */
  if((pOptions = (options*) malloc(sizeof(options))) == NULL)
    {
    printf("Unable to allocate space for options \n");
    exit(1);
    }

  /* Initialize options */

  pOptions->device_name           = NULL;

  pOptions->imm_process_name      = NULL;
  pOptions->seq_len               = 6;
  pOptions->db_seq_len            = 6;

  pOptions->mean_len              = 30;

  pOptions->imm_execve_trace      = FALSE;
  pOptions->imm_threshold         = 0.3;
  pOptions->buffer_size           = BUFFER_LENGTH;
  
  pOptions->log_to_file           = FALSE;
  pOptions->log_path              = NULL;
  pOptions->imm_callback          = 1000;

  pOptions->imm_db_path               = NULL;
  pOptions->fbod                  = FALSE;
  pOptions->fbod_threshold_up        = 0x4fffffff;
   pOptions->fbod_threshold_down       = 0x40000000;

  /* Give the options to the caller */
  return pOptions;
}


void ApplyConfig(int pmFileID, options* pmOptions)
{
  

   if (pmOptions->imm_process_name!=NULL)
     if(ioctl(pmFileID, IMM_SET_NAME, pmOptions->imm_process_name))
       {
	 /* Don't go any further */
	 printf("IMMDaemon: Unable to set process name to trace. \n");
	 exit(1);
       }

   if (pmOptions->imm_execve_trace)
     {
       if(ioctl(pmFileID, IMM_EXECVE_TRACE_ON, 0))
	 {
	   /* Don't go any further */
	   printf("IMMDaemon: Unable to set execve_on. \n");
	   exit(1);
	 }
     }
   
   else
     if(ioctl(pmFileID, IMM_EXECVE_TRACE_OFF, 0))
       {
	 /* Don't go any further */
	 printf("IMMDaemon: Unable to set execve_off. \n");
	 exit(1);
       }
   
}   

int
main (int argc, char *argv[])
{
  GtkWidget *main_window;

  gtk_set_locale ();
  gtk_init (&argc, &argv);

  add_pixmap_directory (PACKAGE_DATA_DIR "/pixmaps");
  add_pixmap_directory (PACKAGE_SOURCE_DIR "/pixmaps");

  gtk_rc_parse ("gtkrc"); 

  gCmap = gdk_colormap_get_system();
  gColorRed.red = 0xffff;
  gColorRed.green = 0;
  gColorRed.blue = 0;
  if (!gdk_color_alloc(gCmap, &gColorRed)) {
    g_error("couldn't allocate color");
  }

  gColorGreen.red = 0;
  gColorGreen.green = 0xdddd;
  gColorGreen.blue = 0;
  if (!gdk_color_alloc(gCmap, &gColorGreen)) {
    g_error("couldn't allocate color");
  }
  
  gOptions = CreateOptions();
  gTree=NewTreeNode();

  setbuf(stdout,NULL);
 
  main_window = create_main_window ();
  gtk_widget_show (main_window);

  gtk_main ();
  return 0;
}

